// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#include <libconfig.h++>
#include <iostream>
#include <stdlib.h>

#include "ApplicationParameters.h"

using namespace std;
using namespace libconfig;


ApplicationParameters::ApplicationParameters(string config_file_name)
{
	Config *myAppConfig = new Config();
	try
	{
		myAppConfig->readFile(config_file_name.c_str());
		cout << "Configuration file loaded ..." << endl;
	}
	catch (FileIOException exc)
	{
		cout << "Configuration file not found!" << endl;
		exit(-1);
	}
	catch (ParseException exc)
	{
		cout << "Error parsing the configuration file!" << endl;
		exit(-1);
	}

	// read & set application settings
	int debug_on = 0;
	if (myAppConfig->lookupValue("application.debugging_on", debug_on))
	{
		if (debug_on == 0)
			this->APP_is_debug_on_ = false;
		else
			this->APP_is_debug_on_ = true;

		if (this->APP_is_debug_on_)
			cout << "Application settings loaded" << endl;
	}
	else
	{
		cout << "Error reading application settings!" << endl;
		exit(-1);
	}

	// read & set data file settings
	int do_sample_weighting = 1;
	if (myAppConfig->lookupValue("data.data_file", this->data_file_) &&
		myAppConfig->lookupValue("data.bag_labels", this->data_baglabels_file_) &&
		myAppConfig->lookupValue("data.sample_labels", this->data_samplelabels_file_) &&
		myAppConfig->lookupValue("data.bag_sample_indices", this->data_bags_to_samples_file_) &&
		myAppConfig->lookupValue("data.train_bag_indices", this->data_trainingset_file_) &&
		myAppConfig->lookupValue("data.test_bag_indices", this->data_testset_file_) &&
		myAppConfig->lookupValue("data.do_sample_weighting", do_sample_weighting))
	{
		if (do_sample_weighting == 0)
			this->data_do_sample_weighting_ = false;
		else
			this->data_do_sample_weighting_ = true;


		if (this->APP_is_debug_on_)
			cout << "Dataset configuration loaded ..." << endl;
	}
	else
	{
		cout << "Error reading dataset configuration!" << endl;
		exit(-1);
	}

	// read & set random forest settings
	int train_sampling;
	if (myAppConfig->lookupValue("randomforest.trees.num_hyperplane_features", this->RF_num_hyperplane_features_) &&
		myAppConfig->lookupValue("randomforest.trees.max_depth", this->RF_max_tree_depth_) &&
		myAppConfig->lookupValue("randomforest.forest_size", this->RF_num_trees_) &&
		myAppConfig->lookupValue("randomforest.trees.num_hypothesis_tries", this->RF_num_hypthesis_tries_) &&
		myAppConfig->lookupValue("randomforest.train_sampling", train_sampling) &&
		myAppConfig->lookupValue("randomforest.trees.min_num_samples_split", this->RF_min_num_samples_split_))
	{
		// Set the bagging type
		switch (train_sampling)
		{
			case 0:
				this->RF_train_sampling_ = BAGGINGWITHREFINE;
				break;
			case 1:
				this->RF_train_sampling_ = BAGGING;
				break;
			case 2:
				this->RF_train_sampling_ = PERMUTATION;
				break;
			case 3:
				this->RF_train_sampling_ = NONE;
				break;
			default:
				this->RF_train_sampling_ = BAGGINGWITHREFINE;
		}

		if (this->APP_is_debug_on_)
			cout << "Random forest configuration loaded ..." << endl;
	}
	else
	{
		cout << "Error loading random forest configuration!" << endl;
		exit(-1);
	}

	// read & set the deterministic annealing parameters
	int da_method;
	int da_loss;
	if (myAppConfig->lookupValue("deterministic_annealing.cooling_parameter_tree", this->DA_approx_cooling_parameter_tree_) &&
		myAppConfig->lookupValue("deterministic_annealing.cooling_parameter_sample", this->DA_approx_cooling_parameter_sample_) &&
		myAppConfig->lookupValue("deterministic_annealing.max_num_epoches", this->DA_num_epoches_) &&
		myAppConfig->lookupValue("deterministic_annealing.cooling_bias_tree", this->DA_approx_cooling_bias_tree_) &&
		myAppConfig->lookupValue("deterministic_annealing.cooling_bias_sample", this->DA_approx_cooling_bias_sample_) &&
		myAppConfig->lookupValue("deterministic_annealing.method", da_method) &&
		myAppConfig->lookupValue("deterministic_annealing.temperature_start", this->DA_exact_temperature_start_) &&
		myAppConfig->lookupValue("deterministic_annealing.temperature_minimum", this->DA_exact_temperature_minimum_) &&
		myAppConfig->lookupValue("deterministic_annealing.cooling_parameter", this->DA_exact_cooling_parameter_) &&
		myAppConfig->lookupValue("deterministic_annealing.loss_function", da_loss))
	{
		switch (da_method)
		{
		case 0:
			this->DA_method_ = EXACT;
			break;
		case 1:
			this->DA_method_ = APPROXIMATION;
			break;
		default:
			this->DA_method_ = APPROXIMATION;
		}

		switch (da_loss)
		{
		case 0:
			this->DA_exact_loss_function_ = HINGE_LOSS;
			break;
		case 1:
			this->DA_exact_loss_function_ = LOGIT_LOSS;
			break;
		case 2:
			this->DA_exact_loss_function_ = ENTROPY;
			break;
		default:
			this->DA_exact_loss_function_ = HINGE_LOSS;
		}

		if (this->APP_is_debug_on_)
			cout << "Determinstic annealing configuration loaded ..." << endl;
	}
	else
	{
		cout << "Error loading deterministic annealing configuration!" << endl;
		exit(-1);
	}
}

ApplicationParameters::~ApplicationParameters()
{
}

ApplicationParameters* ApplicationParameters::instance_ = 0;
string ApplicationParameters::config_file_name = "";
