// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#ifndef APPLICATIONPARAMETERS_H_
#define APPLICATIONPARAMETERS_H_

#include <string>
#include <vector>

using namespace std;


// Random forest training methods
typedef enum
{
	BAGGINGWITHREFINE, BAGGING, PERMUTATION, NONE
} RandomForestTrainSampling;

typedef enum
{
	EXACT, APPROXIMATION
} DeterministicAnnealingMethod;

typedef enum
{
	HINGE_LOSS, LOGIT_LOSS, ENTROPY
} LossFunction;



class ApplicationParameters
{
	private:
		// Singleton
		ApplicationParameters() {}
		ApplicationParameters(string config_file_name);
		ApplicationParameters(const ApplicationParameters&) {}
		static ApplicationParameters* instance_;


		// Application settings
		bool APP_is_debug_on_;

		// Dataset settings
		string data_file_;
		string data_baglabels_file_;
		string data_samplelabels_file_;
		string data_bags_to_samples_file_;
		string data_trainingset_file_;
		string data_testset_file_;
		bool data_do_sample_weighting_;

		// Random forest settings
		int RF_num_trees_;
		int RF_max_tree_depth_;
		int RF_num_hyperplane_features_;
		int RF_num_hypthesis_tries_;
		RandomForestTrainSampling RF_train_sampling_;
		int RF_min_num_samples_split_;

		// Deterministic annealing settings
		DeterministicAnnealingMethod DA_method_;
		int DA_num_epoches_;
		double DA_approx_cooling_parameter_tree_;
		double DA_approx_cooling_parameter_sample_;
		double DA_approx_cooling_bias_tree_;
		double DA_approx_cooling_bias_sample_;
		double DA_exact_temperature_start_;
		double DA_exact_temperature_minimum_;
		double DA_exact_cooling_parameter_;
		LossFunction DA_exact_loss_function_;

	public:
		static ApplicationParameters* instance ()
		{
		   if (!instance_)
			  instance_ = new ApplicationParameters(config_file_name);
		   return instance_;
		}
		static string config_file_name;

		virtual ~ApplicationParameters();


		// Application
		inline bool isDebugOn() const { return this->APP_is_debug_on_; }

		// Dataset
		inline string getDatafile() const { return this->data_file_; }
		inline string getBaglabelsFile() const { return this->data_baglabels_file_; }
		inline string getSamplelabelsFile() const { return this->data_samplelabels_file_; }
		inline string getBagsToSamplesFile() const { return this->data_bags_to_samples_file_; }
		inline string getTrainingSetFile() const { return this->data_trainingset_file_; }
		inline string getTestSetFile() const { return this->data_testset_file_; }
		inline bool getDoSampleWeighting() const { return this->data_do_sample_weighting_; }

		// Random forest
		inline int getRFNumTrees() const { return this->RF_num_trees_; }
		inline int getRFMaxTreeDepth() const { return this->RF_max_tree_depth_; }
		inline int getRFNumHyperplaneFeatures() const { return this->RF_num_hyperplane_features_; }
		inline int getRFNumHyperthesisTries() const { return this->RF_num_hypthesis_tries_; }
		inline RandomForestTrainSampling getRFTrainSamplingMethod() const { return this->RF_train_sampling_; }
		inline int getRFMinNumSamplesForSplit() const { return this->RF_min_num_samples_split_; }

		// Deterministic annealing
		inline DeterministicAnnealingMethod getDAMethod() const { return this->DA_method_; }
		int getDANumEpoches() const { return this->DA_num_epoches_; }
		inline void setDANumEpoches(int new_num_epoches) { this->DA_num_epoches_ = new_num_epoches; }
		inline double getDAApproxCoolingParameterTree() const { return this->DA_approx_cooling_parameter_tree_; }
		inline double getDAApproxCoolingParameterSample() const { return this->DA_approx_cooling_parameter_sample_; }
		inline double getDAApproxCoolingBiasTree() const { return this->DA_approx_cooling_bias_tree_; }
		inline double getDAApproxCoolingBiasSample() const { return this->DA_approx_cooling_bias_sample_; }
		inline double getDAExactTemperatureStart() const { return this->DA_exact_temperature_start_; }
		inline double getDAExactTemperatureMinimum() const { return this->DA_exact_temperature_minimum_; }
		inline double getDAExactCoolingParameter() const { return this->DA_exact_cooling_parameter_; }
		inline LossFunction getDAExactLossFunction() const { return this->DA_exact_loss_function_; }
};

#endif /* APPLICATIONPARAMETERS_H_ */
