// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#include "MILDataset.h"
#include "FileTools.h"


MILDataset::MILDataset()
{
}

MILDataset::~MILDataset()
{
}

void MILDataset::readData( const string& filename_data, const string& filename_samplelabels,
		const string& filename_bag_inst, const string& filename_baglabels )
{
	// Reading the data file
	FileTools file( filename_data );
	MatrixXd data;
	file.read(data);
	num_samples_ = data.rows();
	num_features_ = data.cols();
	num_classes_ = file.getNumClasses();

	// reading the sample labels
	file.setFileName( filename_samplelabels );
	VectorXi sampleLabels;
	file.read(sampleLabels);

	// reading the samples to bag file
	file.setFileName( filename_bag_inst );
	VectorXi samplesToBag;
	file.read(samplesToBag);

	// reading the bag labels
	file.setFileName( filename_baglabels );
	VectorXi bagLabels;
	file.read(bagLabels);
	num_bags_ = bagLabels.size();


	// prepare the "instance to bag" relation
	vector< vector<int> > instancesBagMembership( num_bags_ );
	for ( int nInst = 0; nInst < num_samples_; nInst++ )
	{
		instancesBagMembership[ samplesToBag(nInst) ].push_back(nInst);
	}

	// Create the bags & samples (we also create a vector with all samples to have fast access to the samples for training the random forest)
	bags_.clear();
	samples_.clear();
	MatrixXd sampleDataForBag;
	VectorXi sampleLabelsForBag;
	VectorXd sampleWeightsForBag;

	// we need a global sample index to have a mapping from samples to bags
	// this is needed when testing all samples at a stroke but then using the confidences in the MILForest process
	// CAUTION!!! This cannot be assigned until the train and test datasets are split!
	int global_sample_index = 0;
	VectorXi global_sample_vector;

	int bag_weight = 1;
	VectorXi num_bags_per_class = VectorXi::Zero( num_classes_ );

	// Iterate all bags
	bags_.resize( num_bags_ );
	for ( int b = 0; b < num_bags_; b++ )
	{
		// Find all samples for the bag
		int num_samples_in_bag = (int)instancesBagMembership[b].size();
		sampleDataForBag = MatrixXd( num_samples_in_bag, num_features_ );
		sampleLabelsForBag = VectorXi( num_samples_in_bag );
		sampleWeightsForBag = VectorXd( num_samples_in_bag );

		int sample_index = -1;
		global_sample_vector = VectorXi( num_samples_in_bag );

		for ( int s = 0; s < num_samples_in_bag; s++ )
		{
			// Fill vectors and matrices
			sample_index = instancesBagMembership[b][s];

			for ( int f = 0; f < num_features_; f++ )
				sampleDataForBag(s, f) = data(sample_index, f);
			sampleLabelsForBag(s) = sampleLabels(sample_index);
			sampleWeightsForBag(s) = (double)(1 / (double)num_samples_in_bag);
			global_sample_vector(s) = global_sample_index++;
		}

		// Fill statistics of number of bags per class
		num_bags_per_class( bagLabels(b) )++;

		// Also give the sampleLabelsForBag to set the correct label of the samples
		Bag *current_bag = new Bag(bagLabels(b), bag_weight, sampleDataForBag, sampleLabelsForBag,
				sampleWeightsForBag, global_sample_vector);
		bags_[b] = current_bag;

		for ( int s = 0; s < current_bag->getNumSamples(); s++ )
			samples_.push_back( current_bag->getSamples()[s] );
	}
}

void MILDataset::splitDataset( MILDataset& set1, MILDataset& set2, const string& filename_split_set1, const string& filename_split_set2 )
{
	FileTools file( filename_split_set1 );
	VectorXi set1_bag_indices;
	file.read( set1_bag_indices );
	vector< Bag* > tmp_bags;
	tmp_bags.clear();
	tmp_bags.resize( set1_bag_indices.size() );
	int tmp_global_index = 0;
	for ( int i = 0; i < set1_bag_indices.size(); i++ )
	{
		tmp_bags[i] = bags_[set1_bag_indices(i)];
		for ( int s = 0; s < tmp_bags[i]->getNumSamples(); s++ )
			tmp_bags[i]->getSamples()[s]->setGlobalSampleIndex( tmp_global_index++ );
	}
	set1.setBags( tmp_bags, num_classes_ );

	file.setFileName( filename_split_set2 );
	VectorXi set2_bag_indices;
	file.read( set2_bag_indices );
	tmp_bags.clear();
	tmp_bags.resize( set2_bag_indices.size() );
	tmp_global_index = 0;
	for ( int i = 0; i < set2_bag_indices.size(); i++ )
	{
		tmp_bags[i] = bags_[set2_bag_indices(i)];
		for ( int s = 0; s < tmp_bags[i]->getNumSamples(); s++ )
			tmp_bags[i]->getSamples()[s]->setGlobalSampleIndex( tmp_global_index++ );
	}
	set2.setBags ( tmp_bags, num_classes_ );
}

void MILDataset::setBags( vector<Bag*> bags, int num_classes )
{
	bags_ = bags;
	num_bags_ = (int)bags_.size();
	num_samples_ = 0;
	samples_.clear();
	for ( int b = 0; b < num_bags_; b++ )
	{
		for ( int s = 0; s < bags_[b]->getNumSamples(); s++ )
			samples_.push_back( bags_[b]->getSamples()[s] );
		num_samples_ += bags_[b]->getNumSamples();
	}
	if ( num_bags_ > 0)
		if ( bags_[0]->getNumSamples() > 0 )
			num_features_ = bags_[0]->getSamples()[0]->getNumFeatures();
	num_classes_ = num_classes;
}


const vector< Sample* > MILDataset::getSamples()
{
	int pos_count = 0;
	int neg_count = 0;
	for ( int s = 0; s < (int)samples_.size(); s++ )
	{
		if ( samples_[s]->getLabel() == 1 )
			pos_count++;
		else
			neg_count++;
	}

	for ( int b = 0; b < num_bags_; b++ )
	{
		for ( int s = 0; s < bags_[b]->getNumSamples(); s++ )
		{
			if ( bags_[b]->getSamples()[s]->getLabel() == 1 )
				bags_[b]->getSamples()[s]->setWeihgt( 1.0 / ((double)pos_count * (double)bags_[b]->getNumSamples()) );
			else
				bags_[b]->getSamples()[s]->setWeihgt( 1.0 / ((double)neg_count * (double)bags_[b]->getNumSamples()) );
		}
	}

	return samples_;
}

