// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#include "MILForest.h"


MILForest::MILForest( int num_features, int num_classes ) : rf_( num_features, num_classes )
{
	num_features_ = num_features,
	num_classes_ = num_classes;

	if ( ApplicationParameters::instance()->isDebugOn() )
		cout << "Create the MILForest with " << num_features_ << " features per sample and "
			<< num_classes_ << " classes" << endl;
}

MILForest::~MILForest()
{
}

RandomForestResult MILForest::train( MILDataset& data )
{
	if ( ApplicationParameters::instance()->getDAMethod() == EXACT )
		this->trainExact( data );
	else
		this->trainApprox( data );

	return rf_.test( data.getSamples() );
}

double MILForest::trainAndReturnError( MILDataset& data )
{
	RandomForestResult train_result = this->train( data );

	return train_result.calculateError( data.getSamples() );
}

RandomForestResult MILForest::test( MILDataset& data )
{
	if ( ApplicationParameters::instance()->isDebugOn() )
		cout << endl << "Test the MILForest with " << (int)data.getSamples().size() << " samples" << endl;

	return rf_.test( data.getSamples() );
}

double MILForest::testAndReturnError( MILDataset& data )
{
	RandomForestResult test_result = this->test( data );

	return test_result.calculateError( data.getSamples() );
}




// ######### PRIVATE METHODS ############
void MILForest::trainApprox( MILDataset& data )
{
	if ( ApplicationParameters::instance()->isDebugOn() )
		cout << endl << "Train a MIL forest APPROXIMATION ..." << endl;

	// Train a random forest with all samples and labels according to the bag label
	vector< Sample* > train_samples = data.getSamples();
	if ( ApplicationParameters::instance()->isDebugOn() )
			cout << "Train the initial random forest with " << (int)train_samples.size() << " samples" << endl;

	rf_.train( train_samples );


	RandomForestResult test_result;
	for ( int epoche_value = 0; epoche_value < ApplicationParameters::instance()->getDANumEpoches(); epoche_value++ )
	{
		if ( ApplicationParameters::instance()->isDebugOn() )
			cout << "Epoche " << epoche_value + 1 << " / " <<
				ApplicationParameters::instance()->getDANumEpoches() << endl;

		// Test the random forest to get the confidences for each sample
		if ( ApplicationParameters::instance()->isDebugOn() )
			cout << "\tTest the random forest with " << (int)train_samples.size() << " samples" << endl;
		test_result = rf_.test( train_samples );

		// Some initial stuff
		int num_switches = 0;
		int global_sample_index = -1;
		int previous_label = -1;
		int new_label = -1;

		// Iterate all trees of the forest ...
		for ( int i_tree = 0; i_tree < ApplicationParameters::instance()->getRFNumTrees(); i_tree++ )
		{
			if ( this->doTrainTree( epoche_value ) )
			{
				num_switches = 0;

				// Iterate each bag...
				for ( int b = 0; b < (int)data.getNumBags(); b++ )
				{
					if ( data.getBags()[b]->getBagLabel() == 1 )
					{
						// Only consider positive bags
						bool has_positive = false;

						// Do instance label switching ...
						for ( int s = 0; s < data.getBags()[b]->getNumSamples(); s++ )
						{
							// Get the global index of the current sample in the current bag to be able to get the corresponding /
							// confidence returned from the test function of the random forest!
							global_sample_index = data.getBags()[b]->getSamples()[s]->getGlobalSampleIndex();

							if ( this->doSwitchLabel(epoche_value) )
							{
								previous_label = data.getBags()[b]->getSamples()[s]->getLabel();

								// Dice a new label according to a good probability distribution
								double diced_number = randDouble();
								if (test_result.getConfidences()(global_sample_index, 0) > diced_number)
								{
									data.getBags()[b]->getSamples()[s]->setLabel(0);
								}
								else
								{
									data.getBags()[b]->getSamples()[s]->setLabel(1);
									has_positive = true;
								}

								new_label = data.getBags()[b]->getSamples()[s]->getLabel();
								if (previous_label != new_label)
									num_switches++;
							}
							else
							{
								// Set the label to the prediction of the random forest result
								data.getBags()[b]->getSamples()[s]->setLabel(test_result.getPredictions()(global_sample_index));

								if (test_result.getPredictions()(global_sample_index) == 1)
									has_positive = true;
							}
						}

						// MIL constraint (at least one positive instance in a positive bag)
						if (!has_positive)
						{
							if ( ApplicationParameters::instance()->isDebugOn() )
								cout << "\t\tMIL CONSTRAINT NOT MET!!!!!! in bag " << b + 1 << " / " << (int)data.getNumBags() << endl;

							// Set the instance with highest confidence to label 1
							double best_conf = 0.0;
							int best_sample_index_in_bag = -1;
							for ( int s = 0; s < data.getBags()[b]->getNumSamples(); s++ )
							{
								global_sample_index = data.getBags()[b]->getSamples()[s]->getGlobalSampleIndex();
								if (best_conf < test_result.getConfidences()(global_sample_index, 1))
								{
									best_conf = test_result.getConfidences()(global_sample_index, 1);
									best_sample_index_in_bag = s;
								}
							}
							data.getBags()[b]->getSamples()[best_sample_index_in_bag]->setLabel(1);
						}

					}
					else
					{
						// Do not consider negative bags
					}
				}

				// Retrain the tree
				if ( ApplicationParameters::instance()->isDebugOn() )
					cout << "\tRetrain tree " << i_tree + 1 << "; " << num_switches << " switches;" << endl;

				train_samples = data.getSamples();

				/*
				int pos_count = 0;
				for (int sd = 0; sd < (int)train_samples.size(); sd++)
					if (train_samples[sd]->getLabel() == 1)
						pos_count++;
				cout << "Number of positive examples: " << pos_count << endl;
				*/

				rf_.clearTree( i_tree );
				rf_.trainTree( i_tree, train_samples );
			}
		}
	}
}

void MILForest::trainExact( MILDataset& data )
{
	if ( ApplicationParameters::instance()->isDebugOn() )
			cout << endl << "Train a MIL forest EXACT ..." << endl;

	// Train a random forest with all samples and labels according to the bag label
	vector<Sample*> train_samples = data.getSamples();

	if ( ApplicationParameters::instance()->isDebugOn() )
			cout << "Train the initial random forest with " << (int)train_samples.size() << " samples" << endl;
	rf_.train( train_samples );


	double start_temperature = ApplicationParameters::instance()->getDAExactTemperatureStart();
	double current_temperature = start_temperature;
	double minimal_temperature = ApplicationParameters::instance()->getDAExactTemperatureMinimum();
	int max_epoche = ApplicationParameters::instance()->getDANumEpoches();
	int epoche = 1;
	while( current_temperature > minimal_temperature && epoche <= max_epoche )
	{
		cout << "Current temperature = " << current_temperature << endl;

		// Update the probability estimate of the unlabeled samples
		this->estimateLabelProbabilityDistribution( data, current_temperature );


		// Dice new instance labels for positive bags according to the probability estimate
		this->diceLabelsForPositiveBags( data );
        // Retrain each tree

		#pragma omp parallel for
		for ( int i_tree = 0; i_tree < ApplicationParameters::instance()->getRFNumTrees(); i_tree++ )
		{
			// Retrain the current tree
			//train_samples = this->ds_train_->getAllSamples();
			train_samples = data.getSamples();
			rf_.clearTree( i_tree );
			rf_.trainTree( i_tree, train_samples );
		}

		// Update temperature and epoche...
		current_temperature = this->temperature( epoche,
				ApplicationParameters::instance()->getDAExactCoolingParameter(), start_temperature );
		epoche++;
	}
}


bool MILForest::doTrainTree( int epoche_value )
{
	double p = this->temperatureProbability(epoche_value,
			ApplicationParameters::instance()->getDAApproxCoolingParameterTree());
	p = min( p, 1.0 );
	p = max( p, ApplicationParameters::instance()->getDAApproxCoolingBiasTree() );
	return ( randDouble() < p ) ? true : false;
}

bool MILForest::doSwitchLabel( int epoche_value )
{
	double p = this->temperatureProbability( epoche_value,
			ApplicationParameters::instance()->getDAApproxCoolingParameterSample() );
	p = min( p, 1.0 );
	p = max( p, ApplicationParameters::instance()->getDAApproxCoolingBiasSample() );
	return ( randDouble() < p ) ? true : false;
}

double MILForest::temperatureProbability( int epoche_value, double cooling_parameter )
{
	return exp( (double)epoche_value * cooling_parameter );
}


double MILForest::temperature(int epoche_value, double cooling_parameter, double start_value)
{
	return start_value / pow(cooling_parameter, epoche_value);
}

// EXACT IMPLEMENTATION
void MILForest::estimateLabelProbabilityDistribution( MILDataset& data, double current_temperature)
{
	// Initial stuff
	double p_b_s_0_unnorm = 0.0;
	double p_b_s_1_unnorm = 0.0;
	double normalizing_factor = 0.0;
	double pbs0 = 0.0;
	double pbs1 = 0.0;
	VectorXd label_probs = VectorXd::Zero(2);
	double transformed_confidence = 0.0;

	// Estimate a probability distribution for all samples in a positive labeled bag
	for ( int i_bag = 0; i_bag < data.getNumBags(); i_bag++ )
	{
		if ( data.getBags()[i_bag]->getBagLabel() == 1 )
		{
			// Get the confidences Fk(x_i_j)
			RandomForestResult confidences = rf_.test( data.getBags()[i_bag]->getSamples() );

			// Calculate the probabilities for each sample and each class ... here we only consider the binary class case
			for ( int i_sample = 0; i_sample < data.getBags()[i_bag]->getNumSamples(); i_sample++ )
			{
				// Get the confidence for the positive class ... implicitly also for the negative class (1-conf_pos).
				transformed_confidence = (confidences.getConfidences()(i_sample, 1) - 0.5) * 2.0;
				//transformed_confidence = confidences.getConfidences()(i_sample, 1);

				//cout << "T: " << current_temperature << " conf: " << transformed_confidence << endl;

				p_b_s_0_unnorm = exp( (-this->calculateLoss(transformed_confidence, 0) / current_temperature) - 1.0 );
				p_b_s_1_unnorm = exp( (-this->calculateLoss(transformed_confidence, 1) / current_temperature) - 1.0 );

                //cout << "This is a debug message!!" << endl;

                //p_b_s_0_unnorm = 1.0/(1.0 + exp(-confidences.getConfidences()(i_sample, 0)));
				//p_b_s_1_unnorm = 1.0/(1.0 + exp(-confidences.getConfidences()(i_sample, 1)));

                //cout << "T: " << current_temperature << " conf:" << transformed_confidence << " loss: " << this->calculateLoss(transformed_confidence, 0)
                //        << " loss: " << this->calculateLoss(transformed_confidence, 1) << endl;

                //cout << "T: " << current_temperature << " " << p_b_s_0_unnorm << ", " << p_b_s_1_unnorm << endl;



				normalizing_factor = p_b_s_0_unnorm + p_b_s_1_unnorm;
				pbs0 = p_b_s_0_unnorm / normalizing_factor;
				pbs1 = 1.0 - pbs0;

				//cout << pbs0 << ", " << pbs1 << endl;
                //cout << "===================================" << endl;

				label_probs(0) = pbs0;
				label_probs(1) = pbs1;
				data.getBags()[i_bag]->getSamples()[i_sample]->setLabelProbabilities(label_probs);
			}
		}
	}
}

void MILForest::diceLabelsForPositiveBags( MILDataset& data )
{
	// Iterate all samples and call diceNewLabel method of the samples!
	//#pragma omp parallel for
	for ( int i_bag = 0; i_bag < data.getNumBags(); i_bag++ )
	{
		if ( data.getBags()[i_bag]->getBagLabel() == 1 )
		{
			bool has_positives = false;
			for ( int i_sample = 0; i_sample < data.getBags()[i_bag]->getNumSamples(); i_sample++ )
			{
				data.getBags()[i_bag]->getSamples()[i_sample]->diceNewLabel();
				if ( data.getBags()[i_bag]->getSamples()[i_sample]->getLabel() == 1 )
					has_positives = true;
			}
			// Check the MIL constraint
			if ( !has_positives )
			{
				double best_label_prob = 0.0;
				int best_i_sample = 0;
				for ( int i_sample = 0; i_sample < data.getBags()[i_bag]->getNumSamples(); i_sample++ )
				{
					double label_prob = data.getBags()[i_bag]->getSamples()[i_sample]->getLabelProbabilities()(1);
					if (label_prob > best_label_prob)
					{
						best_label_prob = label_prob;
						best_i_sample = i_sample;
					}
				}
				data.getBags()[i_bag]->getSamples()[best_i_sample]->setLabel(1);
			}
			//if (!has_positives) cout << "Bag has no positive inst" << endl;
		}
	}
}

double MILForest::calculateLoss( double confidence, int label )
{
	if (label == 0)
		label = -1;

	double loss = 0.0;
	switch ( ApplicationParameters::instance()->getDAExactLossFunction() )
	{
	case HINGE_LOSS:
		loss = max( 0.0, 1.0 - (double)label * confidence );
		break;
	case LOGIT_LOSS:
		loss = log( 1.0 + exp( -(double)label * confidence) );
		break;
	case ENTROPY:
		loss -= confidence*log( confidence );
		//loss -= (1.0 - confidence)*log( 1.0 - confidence );
		break;
	default:
		cout << "ERROR CALCULATING THE LOSS FUNCTION" << endl;
	}

	return loss;
}




