// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#include "RandomForest.h"
#include <iostream>
#include <math.h>
#include "Utilities.h"
#include <Eigen/Core>
#include <omp.h>

USING_PART_OF_NAMESPACE_EIGEN


RandomForest::RandomForest( int num_features, int num_classes )
{
	num_features_ = num_features;
	num_classes_ = num_classes;
	num_trees_ = ApplicationParameters::instance()->getRFNumTrees();
	trees_.resize( num_trees_ );
	for ( int i_tree = 0; i_tree < num_trees_; i_tree++ )
	{
		trees_[i_tree] = new RandomTree( num_features_, num_classes_ );
	}
}

RandomForest::~RandomForest()
{
	for (int i_tree = 0; i_tree < num_trees_; i_tree++)
	{
		delete( trees_[i_tree] );
	}
}

double RandomForest::trainAndReturnError( const vector<Sample*> samples )
{
	RandomForestResult result = this->train( samples );
	return result.calculateError( samples );
}

RandomForestResult RandomForest::train( const vector<Sample*> samples )
{
	#pragma omp parallel for
	for ( int i_tree = 0; i_tree < num_trees_; i_tree++ )
	{
		this->trainTree( i_tree, samples );
	}

	return this->test( samples );
}

void RandomForest::trainTree( int tree_index, const vector<Sample*> samples )
{
	switch ( ApplicationParameters::instance()->getRFTrainSamplingMethod() )
	{
	case BAGGINGWITHREFINE:
		this->trainTreeWithBaggingAndRefine( tree_index, samples );
		break;
	case BAGGING:
		this->trainTreeWithBagging( tree_index, samples );
		break;
	case PERMUTATION:
		this->trainTreeWithPermutation( tree_index, samples );
		break;
	case NONE:
		this->trainTreeWithoutBagging( tree_index, samples );
		break;
	}
}

double RandomForest::testAndReturnError( const vector<Sample*> samples, bool show_result )
{
	RandomForestResult result = this->test( samples );
	return result.calculateError( samples, show_result );
}

RandomForestResult RandomForest::test( const vector<Sample*> samples )
{
	MatrixXd confidences = MatrixXd::Zero( (int)samples.size(), num_classes_ );
	RandomForestResult tree_result;
	#pragma omp parallel for
	for ( int t = 0; t < num_trees_; t++ )
	{
		tree_result = trees_[t]->test( samples );
		confidences += tree_result.getConfidences();
	}
	confidences /= num_trees_;

	return RandomForestResult( confidences );
}

void RandomForest::clearTree( int tree_index )
{
	delete( trees_[tree_index] );
	trees_[tree_index] = new RandomTree( num_features_, num_classes_ );
}






// ############### PRIVATE METHODS #######################

void RandomForest::trainTreeWithBagging(int tree_index, const vector<Sample*> samples)
{
	// Subsampling with replacement
	vector<Sample*> inbag_samples, outbag_samples;
	this->subsampleWithReplacement(samples, inbag_samples, outbag_samples);

	trees_[tree_index]->train(inbag_samples);
}

void RandomForest::trainTreeWithBaggingAndRefine(int tree_index, const vector<Sample*> samples)
{
	// Subsampling with replacement
	vector<Sample*> inbag_samples, outbag_samples;
	this->subsampleWithReplacement(samples, inbag_samples, outbag_samples);

	trees_[tree_index]->train(inbag_samples);

	trees_[tree_index]->refine(outbag_samples);
}

void RandomForest::trainTreeWithPermutation(int tree_index, const vector<Sample*> samples)
{
	vector<Sample*> permuted_samples;
	permuted_samples.clear();
	this->permuteSamples(samples, permuted_samples);
	trees_[tree_index]->train(permuted_samples);
}

void RandomForest::trainTreeWithoutBagging(int tree_index, const vector<Sample*> samples)
{
	trees_[tree_index]->train(samples);
}

void RandomForest::subsampleWithReplacement(const vector<Sample*> samples, vector<Sample*>& inbag_samples, vector<Sample*>& outbag_samples)
{
	int num_total_samples = (int)samples.size();

	VectorXi inIndex = VectorXi(num_total_samples);
	VectorXi usedSamples = VectorXi::Zero(num_total_samples);
	inbag_samples.resize( num_total_samples );
	for (int n = 0; n < num_total_samples; n++)
	{
		// get a random index
		inIndex(n) = (int) floor(num_total_samples * randDouble());
		inbag_samples[n] = samples[inIndex(n)];
		usedSamples(inIndex(n)) = 1;
	}
	// get the outbag samples
	for (int n = 0; n < num_total_samples; n++)
	{
		if (!usedSamples(n))
		{
			outbag_samples.push_back(samples[n]);
		}
	}
}

void RandomForest::permuteSamples(const vector<Sample*> samples, vector<Sample*>& permuted_samples)
{
	int num_samples = (int)samples.size();
	VectorXi permutation = randPerm(num_samples);
	permuted_samples.resize( num_samples );
	for (int s = 0; s < num_samples; s++)
	{
		permuted_samples[s] = samples[permutation(s)];
	}
}

