// -*- C++ -*-
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2010 Christian Leistner, leisti@gmail.com
 * Copyright (C) 2010 Christian Leistner, Amir Saffari, Samuel Schulter
 *                    Institute for Computer Graphics and Vision,
 *                    Graz University of Technology, Austria
 */

#include "RandomNode.h"
#include "Utilities.h"
#include <algorithm>

RandomNode::RandomNode( int depth, int num_features, int num_classes )
{
	depth_ = depth;
	is_leaf_ = true;
	num_features_ = num_features;
	num_classes_ = num_classes;
	nodeConf_ = RandomNodeConfidence( num_classes_ );
	random_feature_indices_ = randPerm(num_features, ApplicationParameters::instance()->getRFNumHyperplaneFeatures());
	random_feature_weights_ = VectorXd::Zero(ApplicationParameters::instance()->getRFNumHyperplaneFeatures());
}

RandomNode::~RandomNode()
{
	if ( !is_leaf_ )
	{
		delete( leftChild_ );
		delete( rightChild_ );
	}
}

void RandomNode::train( const vector<Sample*> samples )
{
	if ( this->doSplit(samples) )
	{
		is_leaf_ = false;

		this->findSplittingHypothesis( samples );

		vector<Sample*> leftChildSamples, rightChildSamples;
		this->splitNode( samples, leftChildSamples, rightChildSamples );

		leftChild_ = new RandomNode( depth_ + 1, num_features_, num_classes_ );
		rightChild_ = new RandomNode( depth_ + 1, num_features_, num_classes_ );

		leftChild_->train(leftChildSamples);
		rightChild_->train(rightChildSamples);
	}
	else
	{
		is_leaf_ = true;
		nodeConf_.calculateConfidence( samples );
	}
}

RandomNodeConfidence RandomNode::test( const Sample* sample )
{
	if ( is_leaf_ )
	{
		return nodeConf_;
	}
	else
	{
		double response = sample->getFeatures( random_feature_indices_ ).dot( random_feature_weights_ );
		if (response <= threshold_)
		{
			return leftChild_->test( sample );
		}
		else
		{
			return rightChild_->test( sample );
		}
	}
}

void RandomNode::refine( const vector<Sample*> samples )
{
	if ( is_leaf_ )
	{
		nodeConf_.refineConfidence( samples );
	}
	else
	{
		vector<Sample*> left_node_samples, right_node_samples;
		this->splitNode( samples, left_node_samples, right_node_samples );

		if ( (int)right_node_samples.size() )
		{
			rightChild_->refine( right_node_samples );
		}
		if ( (int)left_node_samples.size() )
		{
			leftChild_->refine( left_node_samples );
		}
	}
}

bool RandomNode::doSplit( const vector<Sample*> samples )
{
	if ( depth_ >= ApplicationParameters::instance()->getRFMaxTreeDepth()
			|| ((int)samples.size()) < ApplicationParameters::instance()->getRFMinNumSamplesForSplit() )
	{
		return false;
	}

	// Test pureness of the node
	int start_label = samples[0]->getLabel();
	for ( int s = 0; s < (int)samples.size(); s++ )
	{
		if ( samples[s]->getLabel() != start_label )
		{
			return true;
		}
	}

	return false;
}

void RandomNode::findSplittingHypothesis( const vector<Sample*> samples )
{
	// Initialize the random_feature_weights_ and threshold_ to some dummy value in case that no hypothesis is found
	fillWithRandomNumbers( this->random_feature_weights_ );
	this->threshold_ = 0.5;

	// Initial declarations
	double bestDGini = 1e10, bestThreshold = 0.0;
	VectorXd tmp_weights = VectorXd::Zero( ApplicationParameters::instance()->getRFNumHyperplaneFeatures() );
	int num_samples = (int)samples.size();
	double sum = 0.0;
	vector< pair<double, int> > responses;
	pair<double, double> cur_gini_thresh;

	// Start searching for hypothesis
	for ( int i_try = 0; i_try < ApplicationParameters::instance()->getRFNumHyperthesisTries(); i_try++ )
	{
		// Dice new random weights
		fillWithRandomNumbers( tmp_weights );

		responses.clear();
		responses.resize( num_samples );
		for (int i_sample = 0; i_sample < num_samples; i_sample++)
		{
			// Calculate the response of the current sample
			sum = samples[i_sample]->getFeatures( random_feature_indices_ ).dot( tmp_weights );

			responses[i_sample] = pair<double, int>( sum, i_sample );
		}

		// Sort the responses
		sort( responses.begin(), responses.end() );

		// Calculate best gini value and according threshold
		cur_gini_thresh = calculateGiniAndThreshold( samples, responses );

		// Max search
		if ( cur_gini_thresh.first < bestDGini )
		{
			bestDGini = cur_gini_thresh.first;
			bestThreshold = cur_gini_thresh.second;
			random_feature_weights_ = tmp_weights;
			threshold_ = cur_gini_thresh.second;
		}
	}
}

pair<double, double> RandomNode::calculateGiniAndThreshold( const vector<Sample*> samples, vector< pair<double, int> > responses )
{
	// Initialize the counters: right takes all at the begining
	double DGini, LGini, RGini, LTotal, RTotal, bestThreshold = 0.0, bestDGini = 1e10;
	vector<double> LCount(num_classes_, 0.0), RCount(num_classes_, 0.0);

	RTotal = 0;
	LTotal = 0;
	// Count the number of samples in each class
	vector< pair<double, int> >::const_iterator resIt(responses.begin()), resEnd(responses.end()), tmpResIt;
	for (; resIt != resEnd; resIt++)
	{
		//int labelIdx = (samples[resIt->second]->getLabel() == 0) ? 0 : 1; // binary class
		int label_idx = samples[resIt->second]->getLabel();
		RCount[label_idx] += samples[resIt->second]->getWeight();
		RTotal += samples[resIt->second]->getWeight();
	}

	// Loop over the sorted values and find the min DGini
	vector< double >::iterator LIt = LCount.begin(), RIt = RCount.begin(), end = LCount.end(), REnd = RCount.end();
	resIt = responses.begin();
	++resIt;
	for (; resIt != resEnd; resIt++)
	{
		tmpResIt = resIt;
		--tmpResIt;
		RTotal -= samples[tmpResIt->second]->getWeight();
		LTotal += samples[tmpResIt->second]->getWeight();

		//int labelIdx = (samples[tmpResIt->second]->getLabel() == 0) ? 0 : 1; // binary class
		int labelIdx = samples[tmpResIt->second]->getLabel();
		RCount[labelIdx] -= samples[tmpResIt->second]->getWeight();
		LCount[labelIdx] += samples[tmpResIt->second]->getWeight();

		if (resIt->first != tmpResIt->first)
		{
			LGini = 0.0;
			RGini = 0.0;
			/*LIt = LCount.begin(); // binary class
			RIt = RCount.begin();
			for (; LIt != end; LIt++, RIt++)      // Calculate Gini index
			{
				LGini += (*LIt/LTotal)*(1.0 - *LIt/LTotal);
				RGini += (*RIt/RTotal)*(1.0 - *RIt/RTotal);
			}*/
			for (int k = 0; k < num_classes_; k++)
			{
				for (int l = 0; l < num_classes_; l++)
				{
					if (k != l)
					{
						LGini += (LCount[k] / LTotal) * (LCount[l] / LTotal);
						RGini += (RCount[k] / RTotal) * (RCount[l] / RTotal);
					}
				}
			}

			DGini = (LTotal*LGini + RTotal*RGini)/(LTotal + RTotal);

			if (DGini < bestDGini)
			{
				bestDGini = DGini;
				bestThreshold = (resIt->first + tmpResIt->first)*0.5;
			}
		}
	}

	return pair<double, double>(bestDGini, bestThreshold);
}


void RandomNode::splitNode( const vector<Sample*> samples, vector<Sample*>& leftNodeSamples, vector<Sample*>& rightNodeSamples )
{
	int num_samples = (int)samples.size();
	double response = 0.0;
	for ( int s = 0; s < num_samples; s++ )
	{
		response = samples[s]->getFeatures( random_feature_indices_ ).dot( random_feature_weights_ );
		if (response <= threshold_)
			leftNodeSamples.push_back( samples[s] );
		else
			rightNodeSamples.push_back( samples[s] );
	}
}


